static const char cvsid[] = "$Revision: 1.3 $";
#include "ConstantsUtils.h"

#ifndef PDFALTO_VERSION_STR
#define PDFALTO_VERSION_STR "unknown"
#endif

namespace ConstantsUtils {
	const char *EXTENSION_XML = ".xml";
	const char *EXTENSION_XML_MAJ = ".XML";
	const char *EXTENSION_VEC = ".vec";
	const char *EXTENSION_SVG = ".svg";
	const char *EXTENSION_JPG = ".jpg";
	const char *EXTENSION_PNG = ".png";
	const char *EXTENSION_PDF = ".pdf";
	const char *EXTENSION_PDF_MAJ = ".PDF";
	const char *NAME_OUTLINE = "outline";
	const char *NAME_METADATA = "metadata";
	const char *NAME_ANNOT = "annot";
	const char *NAME_DATA_DIR = "_data";

	const char *PDFALTO_VERSION = PDFALTO_VERSION_STR;
	const char *PDFALTO_NAME = "pdfalto";

	const char *PDFALTO_CREATOR = "CONTRIBUTORS";

	const char *MEASUREMENT_UNIT = "pixel";
}
