#ifndef CONSTANTSUTILS_H_
#define CONSTANTSUTILS_H_

/** 
 * ConstantsUtils namespace : contain some constants which are used into pdfalto tool <br></br>
 */
namespace ConstantsUtils {
	extern const char *EXTENSION_XML;
	extern const char *EXTENSION_XML_MAJ;
	extern const char *EXTENSION_VEC;
	extern const char *EXTENSION_SVG;
	extern const char *EXTENSION_JPG;
	extern const char *EXTENSION_PNG;
	extern const char *EXTENSION_PDF;
	extern const char *EXTENSION_PDF_MAJ;
	extern const char *NAME_OUTLINE;
	extern const char *NAME_METADATA;
	extern const char *NAME_ANNOT;
	extern const char *NAME_DATA_DIR;

	extern const char *PDFALTO_VERSION;
	extern const char *PDFALTO_NAME;

	extern const char *PDFALTO_CREATOR;

	extern const char *MEASUREMENT_UNIT;
}

#endif
